""" NT User Folder """
__version__='$Revision: 1.2 $'

from AccessControl.User import BasicUser, BasicUserFolder, domainSpecMatch
from AccessControl.PermissionRole import _what_not_even_god_should_do
from Globals import HTMLFile, PersistentMapping
from App.Dialogs import MessageDialog
import win32api
import win32net, win32netcon
import win32security
import time
import string
import re

_re_ntdom_ntname = re.compile(r'(?P<ntdom>[^\\]+)\\(?P<ntname>[^\\]+)')
_DEBUG = 0

def _isTop(self):
    """Return true if self is the toplevel"""
    return self.aq_inner.aq_parent.isTopLevelPrincipiaApplicationObject

def _get_local_ntdom():
    """Returns the local NT Domain
    If the local machine is a domain controller, returns the current domain
    If the local machine is a standalone server, returns the machine name
    """
    infos = win32net.NetServerGetInfo(None, 101)
    type = infos['type']
    if (type & win32netcon.SV_TYPE_DOMAIN_CTRL) or (type & win32netcon.SV_TYPE_DOMAIN_BAKCTRL):
        local_ntdom = win32api.GetDomainName() # XXX - there *has* to be a way to get the domain
        # this computer is a controller for
    else:
        local_ntdom = infos['name']
        
    return string.lower(str(local_ntdom))

class NTUserFolder(BasicUserFolder):
    """ NT User Folder """
    
    meta_type = 'NT User Folder'
    id = 'acl_users'
    title = 'NT User Folder'
    isAUserFolder = 1
    
    manage_options=(
    {'label': 'Local users', 'action': 'manage_local'},
    {'label': 'Users from other NT Domains', 'action': 'manage_ntdomain'},
    {'label': 'Properties', 'action': 'manage_editPropertiesForm'},
    {'label': 'Security', 'action': 'manage_access'},
    {'label': 'Undo',     'action': 'manage_UndoForm'},
    )

    __ac_permissions__=(
        ('Manage users', ('manage_local', 'manage_ntdomain',
            'manage_editProperties', 'manage_editUser', 'manage_users'
            'getUserNames', 'getUserNamesFiltered',
            'getNTDomainUserNamesFiltered',
            'getUsers', 'getUser',
            )),
        )
    
    def _normalize(self, login, default_ntdom=None, check_local=1):
        """ '[ntdom\\]ntname' --> (local, pretty_name, ntdom, ntname)
        local == 1 if user is local
        """
        if not default_ntdom: default_ntdom = self.default_ntdom
        
        login = string.lower(login)
        m = _re_ntdom_ntname.match(login)
        if m:
            d = m.group('ntdom')
            n = m.group('ntname')
        else:
            d = default_ntdom
            n = login
        if check_local and d == self.local_ntdom:
            local = 1
            pn = n
        else:
            local = 0
            pn = d + '\\' + n

        return (local, pn, d, n)

    def __init__(self, default_ntdom, super_user, default_acquire):
        self.local_ntdom = _get_local_ntdom()
        default_ntdom = string.lower(default_ntdom)
        self.default_ntdom = default_ntdom
        self.default_acquire = default_acquire and 1 or 0

        if super_user:
            local, pretty_name, ntdom, ntname = self._normalize(super_user, default_ntdom, 0)
            self._super = Super(pretty_name, ntdom, ntname, [], [], 'Super user', 'This account is mapped to the super user', 0)
            self.super = ntdom, ntname
        else:
            self.super = None

        self.data = PersistentMapping()
    
    def getUserNames(self):
        """ Return a list of user names
        """
        usernames = []
        resume = 0

        # ask the domain controller
        while 1:
            users, junk, resume = win32net.NetUserEnum(None, 0, win32netcon.FILTER_NORMAL_ACCOUNT, resume)
            usernames.extend(map(lambda user, self=self: self.local_ntdom + '\\' + str(user['name']), users))
            if not resume:
                break
        # add those from other domains
        usernames.extend(
            map(lambda user: user[0],
                filter(lambda user: not user[1][0],
                       self.data.items())))
            
        usernames.sort()
        return usernames
    
    def getUsers(self):
        """ Return a list of user objects
        Only users whose settings are not to be acquired are returned
        """
        userobjects = []
        resume = 1

        # ask the domain controller
        while resume:
            users, junk, resume = win32net.NetUserEnum(None, 0, win32netcon.FILTER_NORMAL_ACCOUNT, resume)
            userobjects.extend(map(lambda user, self=self: self.getUser(str(user['name'])), users))
        # add those from others domains (excluding disabled users)
        userobjects.extend(
            map(lambda user, self=self: self.getUser(user[0]),
                filter(lambda user: not user[1][0],
                       self.data.items())))
            
        userobjects = filter(None, userobjects)
        return userobjects
    
    def getUserFrom(self, default_ntdom, login, bypass_acquire=0):
        """ Return the named user object
        """
        isTop = _isTop(self)
        
        local, pretty_name, ntdom, ntname = self._normalize(login, default_ntdom)
        username = ntdom + '\\' + ntname
        
        if isTop and self.super and (ntdom, ntname) == self.super: # check for super user
            return self._super

        if local:
            try:
                userinfo = win32net.NetUserGetInfo(None, ntname, 20)
            
                full_name = str(userinfo['full_name'])
                comment = str(userinfo['comment'])
            except win32api.error:
                return None
        else:
            full_name = comment = '' # XXX - one day, find a way to ask for the controller of
            # the domain, and ask it the full name and comment

        if self.data.has_key(username):
            local, roles, domains, acquire = self.data[username]
        else:
            local = 0
            roles = domains = []
            acquire = self.default_acquire

        if local and acquire \
           and not roles \
           and not bypass_acquire: # acquire only if no roles are defined
            return None # let another user folder deal with this user

        return NTUser(pretty_name, ntdom, ntname, roles, domains, full_name, comment, acquire)

    def getUserLiteral(self, login):
        """ Return the named user object or None if no such user exists
        The user is always returned, even if it is meant to be acquired
        """
        return self.getUserFrom(self.local_ntdom, login, 1)

    def getUser(self, login):
        """ Return the named user object or None if no such user exists or if
        it is to be acquired
        """
        return self.getUserFrom(self.default_ntdom, login)

    manage = manage_main = manage_local = HTMLFile('local', globals())

    manage_ntdomain = HTMLFile('ntdomain', globals())

    def manage_users(self, submit=None, usernames=None, login=None, REQUEST=None):
        """ Add, Remove, ... a ntdomain user
        """
        if submit == 'Add':
            local, pretty_name, ntdom, ntname = self._normalize(login, self.default_ntdom)
            username = ntdom + '\\' + ntname
            if not local:
                if not self.data.has_key(username):
                    data = self.data
                    data[username] = (0, [], [], 0) # ntdomain user don't acquire by default
                    self.data = data
        
        elif submit == 'Delete' and usernames:
            data = self.data
            for username in usernames:
                username = string.lower(username)
                if data.has_key(username):
                    del data[username]
            self.data = data

        if REQUEST: return self.manage_ntdomain(self, REQUEST)
    
    manage_editPropertiesForm = HTMLFile('editPropertiesForm', globals())
    
    def manage_editProperties(self, default_ntdom, REQUEST=None):
        """ Change the default NT Domain
        """
        default_ntdom = string.lower(default_ntdom)
        self.default_ntdom = default_ntdom or self.local_ntdom
        
        if REQUEST: return self.manage_editPropertiesForm(self, REQUEST)

    _editUserForm = HTMLFile('editUserForm', globals())

    def manage_editUserForm(self, REQUEST=None):
        """ Edit user Form
        """
        isTop = _isTop(self)
        return self._editUserForm(self, REQUEST, isTop=isTop)

    def manage_editUser(self, login, acquire=None, domains=None, roles=None, local_user=1, REQUEST=None):
        """ Change the roles and domains of a user
        """
        if domains and not self.domainSpecValidate(domains):
            return MessageDialog(
                title='Illegal value',
                message='Illegal domain specification',
                action ='manage_local')
        local, pretty_name, ntdom, ntname = self._normalize(login, self.default_ntdom)
        username = ntdom + '\\' + ntname
        acquire = acquire and 1 or 0
        data = self.data
        data[username] = (local, roles or [], domains or [], acquire)
        self.data = data
    
        if REQUEST:
            if local_user:
                return self.manage_local(self, REQUEST)
            else:
                return self.manage_ntdomain(self, REQUEST)

    def getUserNamesFiltered(self, pattern=None):
        """ Return a filtered list of *local* user names
        if pattern is not None, only usernames containing pattern are listed
        This method is used by the nt user folder management screen
        """
        usernames = []
        resume = 0
        
        if not pattern:
            while 1:
                users, junk, resume = win32net.NetUserEnum(None, 0, win32netcon.FILTER_NORMAL_ACCOUNT, resume)
                usernames.extend(map(lambda user, self=self: self.local_ntdom + '\\' + str(user['name']), users))
                if not resume:
                    break
        else:
            pattern=string.lower(pattern)
            while 1:
                users, junk, resume = win32net.NetUserEnum(None, 0, win32netcon.FILTER_NORMAL_ACCOUNT, resume)
                usernames.extend(
                    filter(lambda name, pattern=pattern: string.find(string.lower(name), pattern) != -1,
                           map(lambda user, self=self: self.local_ntdom + '\\' + str(user['name']), users)))
                if not resume:
                    break
        usernames.sort()
        return usernames
        
    def getNTDomainUserNamesFiltered(self, pattern=None):
        """ Return a filtered list of *ntdomain* user names
        This method is used by the nt user folder management screen
        """
        if not pattern:
            usernames = map(lambda user: user[0],
                            filter(lambda user: not user[1][0],
                                   self.data.items()))
        else:
            pattern = string.lower(pattern)
            usernames = map(lambda user: user[0],
                            filter(lambda user, pattern=pattern: not user[1][0] and string.find(user[0], pattern) != -1,
                                   self.data.items()))
        usernames.sort()
        return usernames
        
class NTUser(BasicUser):
    """ NT User """
    
    def __init__(self, pretty_name, ntdom, ntname, roles, domains, full_name, comment, acquire):
        self.ntdom, self.ntname = ntdom, ntname
        self.username = ntdom + '\\' + ntname
        self.pretty_name = pretty_name
        
        self.roles = roles
        self.domains = domains

        self.full_name = full_name
        self.comment = comment

        self.acquire = acquire
        
    def getUserName(self):
        """ Return the normalized username 'ntdom\\ntname' of the user
        """
        return self.username

    def getRoles(self):
        """ Return the list of roles assigned to the user
        """
        return self.roles
    
    def getDomains(self):
        """ Return the list of domain restrictions for the user
        """
        return self.domains

    def getPrettyUserName(self):
        """ Return the pretty username of the user.
        If the user is on the local ntdomain, the domain is omitted.
        e.g. if local ntdomain is 'local' and username is 'local\user'
        then pretty_name is 'user'
        """
        return self.pretty_name

    def getFullUserName(self):
        """ Return the full name of the user
        """
        return self.full_name
    
    def getDescription(self):
        """ Return description of the user
        """
        return self.comment

    def isAcquiringSettings(self):
        """ Return true if this user is acquiring settings
        """
        return self.acquire

    def authenticate(self, password, request):
        """ try to authenticate the user
        """
        if self.domains and not domainSpecMatch(self.domains, request):
            return 0

        token = None
        try:
            try:
                token = win32security.LogonUser(
                    self.ntname,
                    self.ntdom,
                    password,
                    win32security.LOGON32_LOGON_NETWORK,
                    win32security.LOGON32_PROVIDER_DEFAULT)
                return 1
            except:
                if _DEBUG:
                    import sys, traceback, string
                    type, val, tb = sys.exc_info()
                    sys.stderr.write(string.join(
                        traceback.format_exception(type, val, tb),''))
                    del type, val, tb
        finally:
            if token:
                win32api.CloseHandle(token)
        return 0

class Super(NTUser):
    """ Super user """
    def allowed(self, parent, roles=None):
        return roles is not _what_not_even_god_should_do

    hasRole=allowed

    def has_role(self, roles, object=None): return 1

    def has_permission(self, permission, object): return 1

_addNTUserFolderForm = HTMLFile('addNTUserFolderForm', globals())

def manage_addNTUserFolderForm(self, REQUEST=None):
    """ Add NT User Folder From
    """
    isTop = _isTop(self)
    local_ntdom = _get_local_ntdom()
    
    return _addNTUserFolderForm(self, REQUEST, local_ntdom=local_ntdom, isTop=isTop)
    
def manage_addNTUserFolder(self, default_ntdom, super_user, default_acquire=None, REQUEST=None):
    """ Add NT User Folder
    """
    ntuf = NTUserFolder(default_ntdom, super_user, default_acquire)
    self = self.this()
    try:
        self._setObject('acl_users', ntuf)
    except:
        return MessageDialog(
            title = 'Item Exists',
            message = 'This object already contains a User Folder',
            action = '%s/manage_main' % REQUEST['URL1'])
    self.__allow_groups__ = ntuf
    
    if REQUEST:
        return self.manage_main(self, REQUEST, update_menu=1)
